try:
    from dicergirl.utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from dicergirl.utils.parser import CommandParser, Commands, Only, Optional, Required
except ImportError:
    from ..utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from ..utils.parser import CommandParser, Commands, Only, Optional, Required

from nonebot.matcher import Matcher
from nonebot.adapters import Bot as Bot
from nonebot.internal.matcher.matcher import Matcher
from nonebot.adapters.onebot.v11 import GroupMessageEvent

hsrcommand = on_startswith(".hsr", priority=1, block=True).handle()

async def hsr_handler(matcher: Matcher, event: GroupMessageEvent):
    """ HSR 车卡指令 """
    if not get_status(event):
        return

    args = format_msg(event.get_message(), begin=".scp", zh_en=True)
    at = get_mentions(event)

    if at and not is_super_user(event):
        return "[Oracle] 权限不足, 无法指定玩家修改人物卡."

    if at:
        qid = at[0]
    else:
        qid = ""

    await matcher.send("[Oracle] 开发中, 请耐心等待...")

commands = {"hsrcommand": "hsr_handler"}


from dicergirl.utils.docimasy import expr
from dicergirl.utils.messages import temporary_madness, madness_end, phobias, manias
from dicergirl.utils.dicer import Dicer
from multilogging import multilogger
from .coccards import coc_cards, coc_attrs_dict
from .investigator import Investigator

import random

logger = multilogger(name="Dicer Girl", payload="COCUtil")

def sc(arg, event):
    """ COC 疯狂检定 """
    reply = []
    try:
        args = arg.split(" ")
        args = list(filter(None, args))
        using_card = False
        s_and_f = args[0].split("/")
        success = Dicer().parse(s_and_f[0])
        success.roll()
        success = success.calc()
        failure = Dicer().parse(s_and_f[1])
        failure.roll()
        failure = failure.calc()
        if len(args) > 1:
            card = {"san": int(args[1]), "name": "未指定调查员"}
            reply.append("[Oracle] 用户指定了应当检定的 SAN 值, 这会使得本次检定不会被记录.")
            using_card = False
        else:
            card = coc_cards.get(event)
            using_card = True
        r = Dicer().roll().calc()
        s = f"[Oracle] 调查员: {card['name']}\n"
        s += f"检定精神状态: {card['san']}\n"
        s += f"理智检定值: {r}, "
        if r <= card["san"]:
            down = success
            s += "检定成功.\n"
        else:
            down = failure
            s += "检定失败.\n"
        s += f"{card['name']} 理智降低了 {down} 点, "
        if down >= card["san"]:
            s += "陷入了永久性疯狂.\n"
        elif down >= (card["san"] // 5):
            s += "陷入了不定性疯狂.\n"
        elif down >= 5:
            s += "陷入了临时性疯狂.\n"
        else:
            s += "未受到严重影响.\n"
        card["san"] -= down
        if card["san"] <= 0:
            card["san"] = 0
        s += f"当前 {card['name']} 的 SAN 值为: {card['san']}"
        reply.append(s)
        if using_card:
            coc_cards.update(event, card)
        return reply
    except:
        return "[Oracle] 产生了未知的错误, 你可以使用`.help sc`指令查看指令使用方法.\n如果你确信这是一个错误, 建议联系开发者获得更多帮助.\n如果你是具有管理员权限, 你可以使用`.debug on`获得更多信息."

def coc_at(event, args):
    """ COC 伤害检定 """
    inv = Investigator().load(coc_cards.get(event))
    method = "+"

    if args:
        d = Dicer().parse(args).roll()
    else:
        d = Dicer().parse("1d6").roll()

    if "d" in inv.db():
        db = Dicer(inv.db()).roll()
        dbtotal = db.outcome
        db = db.db
    else:
        db = int(inv.db())
        dbtotal = db
        if db < 0:
            method = ""

    return f"[Oracle] 投掷 {d.db}{method}{db}=({d.outcome}+{dbtotal})\n造成了 {d.outcome+dbtotal}点 伤害."

def coc_dam(event, args):
    """ COC 承伤检定 """
    card = coc_cards.get(event)
    if not card:
        return "[Oracle] 未找到缓存数据, 请先使用`.coc`指令进行车卡生成角色卡并`.set`进行保存."
    max_hp = card["con"] + card["siz"]
    try:
        arg = int(args[0])
        card["hp"] -= arg
        r = f"[Orcale] {card['name']} 失去了 {arg}点 生命"
    except:
        d = Dicer().parse("1d6").roll()
        card["hp"] -= d.outcome
        r = "[Oracle] 投掷 1D6={d}\n受到了 {d}点 伤害".format(d=d.calc())
    if card["hp"] <= 0:
        card["hp"] = 0
        r += f", 调查员 {card['name']} 已死亡."
    elif (max_hp * 0.8) <= card["hp"] and (card["hp"] < max_hp):
        r += f", 调查员 {card['name']} 具有轻微伤."
    elif (max_hp * 0.6 <= card["hp"]) and (card["hp"] <= max_hp * 0.8):
        r += f", 调查员 {card['name']} 进入轻伤状态."
    elif (max_hp * 0.2 <= card["hp"]) and (card["hp"] <= max_hp * 0.6):
        r += f", 调查员 {card['name']} 身负重伤."
    elif max_hp * 0.2 >= card["hp"]:
        r += f", 调查员 {card['name']} 濒死."
    else:
        r += "."
    coc_cards.update(event, card)
    return r

def coc_ra(event, args):
    """ COC 技能检定 """
    if len(args) == 0:
        return "[Oracle] 错误: 检定技能需要给入技能名称.\n使用`.help ra`指令查看指令使用方法."
    if len(args) > 2:
        return "[Oracle] 错误: 参数过多(最多2需要但%d给予)." % len(args)

    card_data = coc_cards.get(event)
    if not card_data:
        if len(args) == 1:
            return "[Oracle] 你尚未保存人物卡, 请先执行`.coc`车卡并执行`.set`保存.\n如果你希望快速检定, 请执行`.ra [str: 技能名] [int: 技能值]`."

        return str(expr(Dicer(), args[1]))

    inv = Investigator().load(card_data)

    is_base = False
    exp = None
    for _, alias in coc_attrs_dict.items():
        if args[0] in alias:
            exp = int(getattr(inv, alias[0]))
            is_base = True
            break

    if not is_base:
        for skill in inv.skills:
            if args[0] == skill:
                exp = inv.skills[skill]
                break
            else:
                exp = False

    if not exp:
        if len(args) == 1:
            return "[Oracle] 你没有这个技能, 如果你希望快速检定, 请执行`.ra [str: 技能名] [int: 技能值]`."

        if not args[1].isdigit():
            return "[Oracle] 技能值应当为整型数, 使用`.help ra`查看技能检定指令使用帮助."

        return str(expr(Dicer(), int(args[1])))
    elif exp and len(args) > 1:
        if not args[1].isdigit():
            return "[Oracle] 技能值应当为整型数, 使用`.help ra`查看技能检定指令使用帮助."

        reply = [f"[Oracle] 你已经设置了技能 {args[0]} 为 {exp}, 但你指定了检定值, 使用指定检定值作为替代."]
        reply.append(str(expr(Dicer(), int(args[1]))))
        return reply

    time = 1
    r = expr(Dicer(), exp)

    for _ in range(time-1):
        r += expr(Dicer(), exp)

    return r.detail

def ti():
    """ COC 临时疯狂检定 """
    i = random.randint(1, 10)
    r = "临时疯狂判定1D10=%d\n" % i
    r += temporary_madness[i-1]
    if i == 9:
        j = random.randint(1, 100)
        r += "\n恐惧症状为: \n"
        r += phobias[j-1]
    elif i == 10:
        j = random.randint(1, 100)
        r += "\n狂躁症状为: \n"
        r += manias[j-1]
    r += "\n该症状将会持续1D10=%d" % random.randint(1, 10)
    return r

def li():
    """ COC 总结疯狂检定 """
    i = random.randint(1, 10)
    r = "总结疯狂判定1D10=%d\n" % i
    r += madness_end[i-1]
    if i in [2, 3, 6, 9, 10]:
        r += "\n调查员将在1D10=%d小时后醒来" % random.randint(1, 10)
    if i == 9:
        j = random.randint(1, 100)
        r += "\n恐惧症状为: \n"
        r += phobias[j-1]
    elif i == 10:
        j = random.randint(1, 100)
        r += "\n狂躁症状为: \n"
        r += manias[j-1]
    return r

def coc_en(event, args):
    """ COC 技能成长检定 """
    if not args:
        return "[Oracle] 错误: 检定技能需要给入技能名称.\n使用`.help ra`指令查看指令使用方法."

    try:
        arg = int(args[1])
    except ValueError:
        return "[Oracle] 错误: 给定需要消耗的激励点应当为整型数.\n使用`.help ra`指令查看指令使用方法."

    check = random.randint(1, 100)

    if check > arg or check > 95:
        plus = random.randint(1, 10)
        r = "判定值%d, 判定成功, 技能成长%d+%d=%d" % (check, arg, plus, arg+plus)
        return r + "\n温馨提示: 如果技能提高到90%或更高, 增加2D6理智点数。"
    else:
        return "判定值%d, 判定失败, 技能无成长。" % check


from typing import Dict, List, Any
try:
    from .errors import NoneTypeCommandError, CommandRequired, TooManyAliasCommandError
except ImportError:
    from errors import NoneTypeCommandError, CommandRequired, TooManyAliasCommandError

class Optional:
    """ 可选指令 """
    def __init__(self, key: str, cls: type, default: Any=None):
        if not key:
            raise NoneTypeCommandError("Optional parameter must not be `None`.")

        if isinstance(key, str):
            key = [key, ]

        self.key = key
        self.cls = cls
        self.default = default

    def __str__(self):
        return self.key[0]

class Required:
    """ 必选指令 """
    def __init__(self, key, cls: type, default: Any=None):
        if not key:
            raise NoneTypeCommandError("Required parameter must not be `None`.")

        if isinstance(key, str):
            key = [key, ]

        self.key = key
        self.cls = cls
        self.default = default

    def __str__(self):
        return self.key[0]

class Only:
    """ 布尔指令 """
    def __init__(self, key, default: bool=None):
        if not key:
            raise NoneTypeCommandError("Bool parameter must not be `None`.")

        if isinstance(key, str):
            key = [key, ]

        self.key = key
        self.default = default

    def __str__(self):
        return self.key[0]

class Commands(list):
    """ 指令集合 """
    def __init__(self, *args, **kwargs):
        super(Commands, self).__init__(*args, **kwargs)

    def __required__(self) -> List[Required]:
        return [required for required in self if isinstance(required, Required)]

    def __optional__(self) -> List[Optional]:
        return [optional for optional in self if isinstance(optional, Optional)]

    def get_plain_required(self) -> List[str]:
        return [str(required) for required in self if isinstance(required, Required)]

    def get_plain_optional(self) -> List[str]:
        return [str(optional) for optional in self if isinstance(optional, Optional)]

    def get_plain_commands(self) -> List[str]:
        return [str(command) for command in self]

def required(commands: Commands):
    return commands.__required__()

def optional(commands: Commands):
    return commands.__optional__()

class CommandParser:
    """指令解析类
    示例:
        ```python
        cp = CommandParser(
            Commands([Optional("optional"), Required("required"), Only("bool"), ...]),
            args = ["required", "valueofrequired", "bool"],
            auto = True # 自动解析指令, `auto=True`时, `args`不得为空
            ).results
        print(cp["optional"]) # 输出为`None`
        print(cp["required"]) # 输出为`valueofrequired`
        print(cp["bool"]) # 输出为`True`
        ```
    """
    def __init__(self, commands: Commands=None, args: List[str]=None, auto: bool=False):
        self.results: Dict[str, str] = {}

        if not isinstance(commands, Commands):
            raise TypeError("指令槽必须为类`Commands`.")
        if args and not isinstance(args, (list, tuple)):
            raise TypeError("参数槽必须为列或数组.")

        self.commands = commands
        self.args = args
        self.nothing = False

        if auto:
            self.shlex()

    def shlex(self, args: List[str]=None):
        """ 开始拆析指令集合 """
        if not args:
            args = self.args

        if not isinstance(args, (list, tuple)):
            raise TypeError("指令切片必须传入列或数组.")

        results: Dict[str, str] = {}
        nothing: bool = True

        for command in self.commands:
            if isinstance(command, Only):
                if set(command.key) & set(args):
                    results[command.key[0]] = True
                    nothing = False
                else:
                    results[command.key[0]] = False
                continue

            key = list(set(command.key) & set(args))
            if len(key) > 1:
                raise TooManyAliasCommandError("Too many alias parameters.")

            if key:
                index = args.index(key[0])
                if len(args) > index + 1:
                    try:
                        value = command.cls(args[index+1])
                    except ValueError:
                        raise TypeError(f"Value type of {command.key} is mismatch, {command.key} required but {type(args[index+1])} was given.")
                    results[command.key[0]] = value
                    nothing = False
                else:
                    results[command.key[0]] = command.default
            else:
                if isinstance(command, Required):
                    raise CommandRequired(f"Required parameter `{command.key[0]}` not found.")

                results[command.key[0]] = command.default

        self.results = results
        self.nothing = nothing

    def __iter__(self):
        return iter(self.results.items())

if __name__ == "__main__":
    cp = CommandParser(
        Commands([Only("cache"), Optional("age", int), Optional(("name", "n"), str, "欧若可"), Optional("sex", str), Optional("roll", int)]),
        )
    cp.args = ["cache", "age", "20", "n", "先生"]
    cp.shlex()
    print(cp.results)
    cp = CommandParser(
        Commands([Only("cache"), Required("test", int)])
    )
    cp.args = ["cache", "test", "222"]
    cp.shlex()
    print(dict(cp))


from .adventurer import Adventurer
from .dndcards import dnd_cache_cards

try:
    from dicergirl.utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from dicergirl.utils.parser import CommandParser, Commands, Only, Optional, Required
except ImportError:
    from ..utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from ..utils.parser import CommandParser, Commands, Only, Optional, Required

from nonebot.matcher import Matcher
from nonebot.adapters import Bot as Bot
from nonebot.adapters.onebot.v11 import Bot as V11Bot
from nonebot.internal.matcher.matcher import Matcher
from nonebot.adapters.onebot.v11 import GroupMessageEvent

dndcommand = on_startswith(".dnd", priority=1, block=True).handle()

async def dnd_handler(matcher: Matcher, event: GroupMessageEvent):
    """ DND 车卡指令 """
    if not get_status(event) and not event.to_me:
        return

    args = format_msg(event.get_message(), begin=".dnd")
    if len(args) > 1:
        await matcher.send("[Oracle] 错误: 参数超出预计(1需要 但 %d传入), 指令驳回." % len(args))
        return True

    try:
        if len(args) == 0:
            raise ValueError
        args = int(args[0])
    except ValueError:
        await matcher.send(f'警告: 参数 {args} 不合法, 使用默认值 20 替代.')
        args = 20

    adv = Adventurer()
    adv.age_check(args)
    adv.init()

    if adv.int[0] <= 8:
        await matcher.send("[Orcale] 很遗憾, 检定新的冒险者智力不足, 弱智是不允许成为冒险者的, 请重新进行车卡检定.")
        return True

    if 15 <= args and args < 90:
        dnd_cache_cards.update(event, adv.__dict__, save=False)
        await matcher.send(str(adv.output()))
    return True

commands = {"dndcommand": "dnd_handler"}


from .adventurer import Adventurer
from .nbhandlers import commands
from .dndcards import dnd_cards
from .dndutils import dnd_at, dnd_dam, dnd_ra

dnd_cards.load()

__type__ = "plugin"
__charactor__ = Adventurer
__name__ = "dnd"
__cname__ = "冒险者"
__nbhandler__ = nbhandlers
__nbcommands__ = commands
__commands__ = {
    "at": dnd_at,
    "dam": dnd_dam,
    "ra": dnd_ra,
}
__description__ = "DND 模式是以游戏 龙与地下城(Dungeons & Dragons) 为背景的 TRPG 跑团模式."


DEFAULT_DURABILITY = 10000
GUN_DEFAULT_DURABILITY = 1000000

scp_attrs_dict = {
    "名字": ["name", "名字", "名称", "姓名"],
    "性别": ["sex", "性别"],
    "年龄": ["age", "年龄"],
    "强度": ["str", "强度", "力量", "攻击"],
    "健康": ["hth", "健康", "体质"],
    "感知": ["per", "感知", "灵感"],
    "灵巧": ["dex", "灵巧", "敏捷"],
    "命运": ["fte", "命运", "幸运", "气运", "运气"],
    "魅力": ["chr", "魅力", "外貌"],
    "情报": ["int", "情报", "智力", "侦查"],
    "意志": ["wil", "意志", "精神", "理智"],
    "生命": ["hp", "生命"],
    "声望": ["rep", "声望", "信誉"],
    "余额": ["money", "金钱", "余额", "金额"],
    "等级": ["level", "等级"]
}
""" SCP 基础属性 """

knowledge_data = {
    "解剖": ["解剖", "急救"],
    "古代语言": ["古代语言"],
    "建筑": ["建筑"],
    "天文学": ["天文学"],
    "护理": ["护理", "养育"],
    "计算机": ["计算机", "黑客攻击"],
    "烹饪": ["烹饪"],
    "拆除": ["拆除"],
    "时尚": ["时尚", "礼仪"],
    "赌博": ["赌博", "游戏"],
    "一般知识": ["一般知识"],
    "历史": ["历史", "知识"],
    "识别味道": ["识别味道", "气味"],
    "调查": ["调查"],
    "法律": ["法律", "政治"],
    "数学": ["数学"],
    "机械": ["机械"],
    "导航": ["导航"],
    "神秘": ["神秘", "SCP 知识"],
    "心理学": ["心理学"],
    "宗教": ["宗教"],
    "研究": ["研究", "互联网"],
    "科学": ["科学", "物理"],
    "生存": ["生存", "追踪"],
    "技术": ["技术"],
    "视觉信号": ["视觉信号", "手势语言"],
}
""" SCP 知识 """

skills_data = {
    "近战": ["近战"],
    "突击步枪": ["突击步枪"],
    "手枪": ["手枪"],
    "重型火炮": ["重型火炮"],
    "狙击": ["猎枪", "狙击", "步枪"],
    "霰弹枪": ["霰弹枪"],
    "冲锋枪": ["冲锋枪"],
    "演戏": ["演戏", "说谎"],
    "艺术": ["艺术", "音乐"],
    "挣脱": ["挣脱", "逃脱"],
    "抛投": ["抛投", "接"],
    "攀爬": ["攀爬"],
    "伪装": ["伪装", "混合"],
    "驾驶": ["驾驶"],
    "开锁": ["开锁"],
    "扒手": ["扒手"],
    "飞行员": ["飞行员"],
    "潜行": ["潜行", "隐藏"],
    "表演技巧": ["表演技巧"],
    "游泳": ["游泳"],
    "教导": ["教导"],
    "摔跤": ["摔跤", "缴械"],
    "写作": ["写作"],
}
""" SCP 技能 """

ability_data = {
    "意识": ["意识", "反应"],
    "闪避": ["闪避", "格挡"],
    "动物驯服": ["动物驯服"],
    "运动": ["运动"],
    "同理心": ["同理心"],
    "耐力": ["耐力"],
    "主动权": ["主动权"],
    "威吓": ["威吓", "嘲弄"],
    "直觉": ["直觉"],
    "跳跃": ["跳跃"],
    "领导能力": ["领导能力"],
    "提拉力量": ["提拉力量"],
    "谈判": ["谈判", "说服"],
    "倒地": ["倒地", "死亡抗性"],
    "疼痛抗性": ["疼痛抗性"],
    "自我控制": ["自我控制"],
}
""" SCP 能力 """

weapons = {
    1: {
        "泰瑟枪": {
            "base": "1d4",
            "x": "0",
            "price": 210,
            "durability": DEFAULT_DURABILITY,
            "using": "remote"
        },
        "弩": {
            "base": "1d10",
            "x": "3d8",
            "price": 48,
            "durability": DEFAULT_DURABILITY,
            "using": "remote"
        },
        "鲁格手枪": {
            "base": "1d6",
            "x": "1d6",
            "price": 480,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "M206": {
            "base": "2d6",
            "x": "1d6",
            "price": 360,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "RM380": {
            "base": "1d8",
            "x": "1d4",
            "price": 360,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "沃尔特-RPK": {
            "base": "1d6",
            "x": "1d8",
            "price": 900,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "棒球棍": {
            "base": "0",
            "x": "2d8",
            "price": 24,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "指虎": {
            "base": "1d6",
            "x": "1d8",
            "price": 25,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "撬棍": {
            "base": "1d12",
            "x": "1d10",
            "price": 96,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "甩棍": {
            "base": "1d8",
            "x": "1d4",
            "price": 510,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "生存小刀": {
            "base": "1d8",
            "x": "1d6",
            "price": 120,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "鞭子": {
            "base": "1d10",
            "x": "1d6",
            "price": 168,
            "durability": DEFAULT_DURABILITY,
            "using": "infight"
        },
        "燃烧瓶": {
            "base": "4d10",
            "x": "0",
            "price": 96,
            "durability": 1,
            "using": "throw"
        }
    },
    2: {
        "克洛克-19": {
            "base": "2d6",
            "x": "1d8+3",
            "price": 750,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "M1911": {
            "base": "2d8",
            "x": "1d10+3",
            "price": 1500,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "M4A1": {
            "base": "4d8",
            "x": "2d6",
            "price": 4500,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "马林-336C": {
            "base": "2d8",
            "x": "2d8+2",
            "price": 750,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "和平使者": {
            "base": "2d8",
            "x": "1d8",
            "price": 1050,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        },
        "SAKO-Finnlight": {
            "base": "2d8",
            "x": "3d8",
            "price": 1650,
            "durability": GUN_DEFAULT_DURABILITY,
            "using": "remote"
        }
    }
}
""" SCP 武器 """

all_names = list(set().union(scp_attrs_dict, knowledge_data, skills_data, ability_data))
all_alias = []
all_alias_dict = {}

for alias in dict(scp_attrs_dict, **knowledge_data, **skills_data, **ability_data).values():
    all_alias += alias

for key, alias in dict(scp_attrs_dict, **knowledge_data, **skills_data, **ability_data).items():
    for alia in alias:
        all_alias_dict[alia] = key


from .agent import Agent
from .nbhandlers import commands
from .scpcards import scp_cards, scp_cache_cards
from .attributes import scp_attrs_dict
from .scputils import scp_at, scp_dam, scp_ra, scp_en

scp_cards.load()

__type__ = "plugin"
__charactor__ = Agent
__name__ = "scp"
__cname__ = "特工"
__cards__ = scp_cards
__cache__ = scp_cache_cards
__nbhandler__ = nbhandlers
__nbcommands__ = commands
__commands__ = {
    "at": scp_at,
    "dam": scp_dam,
    "ra": scp_ra,
    "en": scp_en
}
__baseattrs__ = scp_attrs_dict
__description__ = "SCP 模式是基于SCP基金会(SCP Foundation) 设定的 TRPG 跑团模式."


from .investigator import Investigator
from .coccards import coc_cards, coc_cache_cards, coc_rolls
from .cocutils import sc, ti, li

try:
    from ..utils.utils import format_msg, get_status, on_startswith, format_str
    from ..utils.parser import CommandParser, Commands, Only, Optional, Required
except ImportError:
    from dicergirl.utils.utils import format_msg, get_status, on_startswith, format_str
    from dicergirl.utils.parser import CommandParser, Commands, Only, Optional, Required

from nonebot.matcher import Matcher
from nonebot.adapters import Bot as Bot
from nonebot.adapters.onebot.v11 import Bot as V11Bot
from nonebot.internal.matcher.matcher import Matcher
from nonebot.adapters.onebot.v11 import MessageEvent, GroupMessageEvent

coccommand = on_startswith(".coc", priority=1, block=True).handle()
ticommand = on_startswith(".ti", priority=2, block=True).handle()
licommand = on_startswith(".li", priority=2, block=True).handle()
sccommand = on_startswith(".sc", priority=2, block=True).handle()

async def coc_handler(matcher: Matcher, event: GroupMessageEvent):
    """ COC 车卡指令 """
    if not get_status(event) and not event.to_me:
        return

    args = format_msg(event.get_message(), begin=".coc", zh_en=True)
    qid = event.get_user_id()
    commands = CommandParser(
        Commands([
            Only("cache", False),
            Optional("set", int),
            Optional("age", int, 20),
            Optional("name", str),
            Optional("sex", str, "女"),
            Optional("roll", int, 1)
            ]),
        args=args,
        auto=True
        ).results
    toroll = commands["roll"]

    if commands["set"]:
        coc_cards.update(event, coc_rolls[qid][commands["set"]], save=True)
        inv = Investigator().load(coc_rolls[qid][commands["set"]])
        await matcher.send(f"使用序列 {commands['set']} 卡:\n{inv.output()}")
        coc_rolls[qid] = {}
        return

    age = commands["age"]
    name = commands["name"]

    if not (15 <= age and age < 90):
        await matcher.send(Investigator().age_change(age))
        return

    reply = ""
    if qid in coc_rolls.keys():
        rolled = len(coc_rolls[qid].keys())
    else:
        coc_rolls[qid] = {}
        rolled = 0

    for i in range(toroll):
        inv = Investigator()
        inv.age_change(age)
        inv.sex = commands["sex"]

        if name:
            inv.name = name

        coc_rolls[qid][rolled+i] = inv.__dict__
        count = inv.rollcount()
        reply += f"天命编号: {rolled+i}\n"
        reply += inv.output() + "\n"
        reply += f"共计: {count[0]}/{count[1]}\n"

    if toroll == 1:
        coc_cache_cards.update(event, inv.__dict__, save=False)

    reply.rstrip("\n")
    await matcher.send(reply)

async def ticommandhandler(matcher: Matcher, event: MessageEvent):
    """ COC 临时疯狂检定指令 """
    if not get_status(event):
        return

    try:
        await matcher.send(ti())
    except:
        await matcher.send("[Oracle] 未知错误, 执行`.debug on`获得更多信息.")

async def licommandhandler(matcher: Matcher, event: MessageEvent):
    """ COC 总结疯狂检定指令 """
    if not get_status(event):
        return

    try:
        await matcher.send(li())
    except:
        await matcher.send("[Oracle] 未知错误, 执行`.debug on`获得更多信息.")

async def sccommandhandler(matcher: Matcher, event: GroupMessageEvent):
    """ COC 疯狂检定指令 """
    if not get_status(event):
        return

    args = format_str(event.get_message(), begin=".sc")
    scrs = sc(args, event)

    if isinstance(scrs, list):
        for scr in scrs:
            await matcher.send(scr)
    else:
        await matcher.send(scrs)

commands = {
    "coccommand": "coc_handler",
    "ticommand": "ticommandhandler",
    "licommand": "licommandhandler",
    "sccommand": "sccommandhandler"
    }


from pathlib import Path
from typing import Union, Dict, List, Callable
from loguru._logger import Logger
from nonebot.consts import STARTSWITH_KEY
from nonebot.plugin import on_message
from nonebot.rule import Rule
from nonebot.matcher import Matcher
from pypi_simple import PyPISimple

import json
import uuid
import re
import inspect
import json
import asyncio
import httpx

try:
    from .decorators import translate_punctuation
    from .settings import get_package, setconfig, getconfig, change_status, load_status_settings
    from .multilogging import multilogger
except ImportError:
    from dicergirl.utils.decorators import translate_punctuation
    from dicergirl.utils.settings import get_package, setconfig, getconfig, change_status, load_status_settings
    from dicergirl.utils.multilogging import multilogger

package = get_package()
""" 当前 Dicer Girl 运行平台 """
version = "3.2.8"
""" Dicer Girl 版本号 """
current_dir = Path(__file__).resolve().parent
""" Dicer Girl 当前目录 """
dicer_girl_dir = Path.home() / ".dicergirl"
data_dir = dicer_girl_dir / "data"
log_dir = dicer_girl_dir / "log"
_dicer_girl_status = data_dir / "status.json"
_super_user = data_dir / "super_user.json"
_loggers_cachepath = data_dir / "loggers.json"
_modes_cachepath = data_dir / "modes.json"
logger = multilogger(name="Dicer Girl", payload="utils")
""" `utils.py`日志系统 """
su_uuid = (str(uuid.uuid1()) + str(uuid.uuid4())).replace("-", "")
loggers: Dict[str, Dict[int, List[Logger | str]]] = {}
""" 正在运行的日志 """
saved_loggers: Dict[str, dict]
""" 存储的日志 """

try:
    from nonebot.adapters.onebot.v11 import MessageEvent, GroupMessageEvent
except ModuleNotFoundError:
    logger.warning("未找到依赖`Nonebot2`, 请检查你的配置.")
    class MessageEvent:
        pass
    class GroupMessageEvent:
        pass

def init() -> None:
    """ 骰娘初始化 """
    global saved_loggers
    dirs: Dict[str, List[Path, list]] = {
        "Dicer Girl": [dicer_girl_dir, "dir"],
        "Dicer Girl 数据": [data_dir, "dir"],
        "Dicer Girl 日志": [log_dir, "dir"],
        "Dicer Girl 状态管理": [_dicer_girl_status, "file"],
        "日志管理": [_loggers_cachepath, "file"],
        "跑团模式存储": [_modes_cachepath, "file"],
        "超级用户存储": [_super_user, "file"]
    }
    for name, dir in dirs.items():
        if not dir[0].exists():
            logger.info(f"{name}{'文件夹' if dir[1] == 'dir' else '文件'}未建立, 建立它.")
            if dir[1] == "dir":
                dir[0].mkdir()
            else:
                with open(dir[0], "w", encoding="utf-8") as f:
                    f.write("{}")
    saved_loggers = load_loggers()
    load_status()

class StartswithRule:
    """自定义的指令检查方法
    允许:
        1. 无视中英文字符串
        2. 无视前后多余空字符
    """
    __slots__ = ("msg", "ignorecase")

    def __init__(self, msg, ignorecase=False):
        self.msg = msg
        self.ignorecase = ignorecase

    def __repr__(self) -> str:
        return f"Startswith(msg={self.msg}, ignorecase={self.ignorecase})"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, StartswithRule)
            and frozenset(self.msg) == frozenset(other.msg)
            and self.ignorecase == other.ignorecase
        )

    def __hash__(self) -> int:
        return hash((frozenset(self.msg), self.ignorecase))

    async def __call__(self, event, state) -> bool:
        try:
            text = translate_punctuation(event.get_plaintext()).strip()
        except Exception:
            return False
        if match := re.match(
            f"^(?:{'|'.join(re.escape(prefix) for prefix in self.msg)})",
            text,
            re.IGNORECASE if self.ignorecase else 0,
        ):
            state[STARTSWITH_KEY] = match.group()
            return True
        return False

def startswith(msg, ignorecase=True) -> Rule:
    """ 实例化指令检查方法 """
    if isinstance(msg, str):
        msg = (msg,)

    return Rule(StartswithRule(msg, ignorecase))

def on_startswith(commands, priority=0, block=True) -> Matcher:
    """ 获得`Nonebot2`指令检查及参数注入方法 """
    if isinstance(commands, str):
        commands = (commands, )

    return on_message(startswith(commands, True), priority=priority, block=block, _depth=1)

def load_modes() -> Dict[str, list]:
    """ 加载当前不同群聊的跑团模式 """
    return json.loads(open(_modes_cachepath, "r").read())

def set_mode(event, mode) -> bool:
    """ 设置当前群聊的跑团模式 """
    lm = load_modes()
    lm[get_group_id(event)] = mode
    json.dump(lm, open(_modes_cachepath, "w"))

def get_mode(event) -> str:
    """ 获得当前群聊的跑团模式 """
    lm = load_modes()
    if not get_group_id(event) in lm.keys():
        lm[get_group_id(event)] = "scp"
        json.dump(lm, open(_modes_cachepath, "w"))
        return "scp"

    return lm[get_group_id(event)]

def load_loggers() -> Dict[str, list]:
    """ 加载所有的已存储的日志 """
    return json.loads(open(_loggers_cachepath, "r").read())

def get_loggers(event) -> List[str]:
    """ 获取`event`所指向的群聊中所有的日志 """
    got_loggers = json.load(open(_loggers_cachepath, "r"))
    if not get_group_id(event) in got_loggers:
        return []

    return got_loggers[get_group_id(event)]

def add_logger(event: GroupMessageEvent, logname) -> bool:
    """ 新增日志序列 """
    global saved_loggers
    if not get_group_id(event) in saved_loggers.keys():
        saved_loggers[get_group_id(event)] = []

    try:
        saved_loggers[get_group_id(event)].append(logname)
        json.dump(saved_loggers, open(_loggers_cachepath, "w"))
        return True
    except:
        return False

def remove_logger(event: GroupMessageEvent, id: int) -> Dict[str, list]:
    """ 从存储的`loggers.json`中移除指定`logger` """
    saved_loggers[get_group_id(event)].pop(id)
    json.dump(saved_loggers, open(_loggers_cachepath, "w"))
    return saved_loggers

def set_config(appid, token) -> dict:
    """ 在`QQGuild`模式中设置频道机器人`appid`以及`token`. """
    return setconfig(appid, token, path=dicer_girl_dir, filename="config.yaml")

def get_config() -> dict:
    """ 获取`QQGuild`模式中频道机器人的`appid`以及`token`. """
    return getconfig(path=dicer_girl_dir, filename="config.yaml")

def format_msg(message, begin=None, zh_en=False) -> List[str]:
    """ 骰娘指令拆析为`list`的方法 """
    msg = format_str(message, begin=begin).split(" ")
    outer = []
    regex = r'(\d+)|([a-zA-Z\u4e00-\u9fa5]+)' if not zh_en else r"(\d+)|([a-zA-Z]+)|([\u4e00-\u9fa5]+)"

    for m in msg:
        m = re.split(regex, m)
        m = list(filter(None, m))
        outer += m

    msg = outer
    msg = list(filter(None, msg))
    logger.debug(msg)
    return msg

def format_str(message: str, begin=None) -> str:
    """ 骰娘指令转义及解析 """
    regex = r"[<\[](.*?)[\]>]"
    msg = re.sub("\s+", " ", re.sub(regex, "", str(message).lower())).strip(" ")
    msg = translate_punctuation(msg)
    logger.debug(msg)

    if begin:
        if isinstance(begin, str):
            begin = [begin, ]
        elif isinstance(begin, tuple):
            begin = list(begin)

        begin.sort(reverse=True)
        for b in begin:
            msg = msg.replace(b, "").lstrip(" ")

    logger.debug(msg)
    return msg

def get_mentions(event: GroupMessageEvent) -> List[str]:
    """ 获取`event`指向的消息所有被`@`的用户 QQ 号 """
    mentions = []
    message = json.loads(event.json())["message"]

    for mention in message:
        if mention["type"] == "at":
            mentions.append(mention["data"]["qq"])

    return mentions

def get_handlers(main) -> List[Callable]:
    """ 获取目前所有的指令触发函数方法 """
    commands_functions = []

    for _, obj in vars(main).items():
        if inspect.isfunction(obj) and hasattr(obj, '__annotations__'):
            annotations = obj.__annotations__
            if annotations.get('message') is GroupMessageEvent:
                commands_functions.append(obj)

    return commands_functions

def get_group_id(event) -> str:
    """ 获取`event`指向的群聊`ID` """
    try:
        if package == "qqguild":
            return str(event.channel_id)
        elif package == "nonebot2":
            return str(event.group_id)
    except:
        logger.warning(f"超出预计的 package: {package}, 将 Group ID 设置为 0.")
        return "0"

def get_user_id(event) -> str:
    """ 获取`event`指向的用户`ID` """
    try:
        if package == "qqguild":
            return eval(str(event.author))["id"]
        elif package == "nonebot2":
            return str(event.get_user_id())
    except:
        logger.warning(f"超出预计的 package: {package}, 将 User ID 设置为 0.")
        return "0"

def get_user_card(event) -> str:
    """ 获取`event`指向的用户群名片 """
    try:
        raw_json = json.loads(event.json())['sender']
        if raw_json['card']:
            return raw_json['card']
        else:
            return raw_json['nickname']
    except:
        return "未知用户"

def add_super_user(message) -> bool:
    """ 新增超级管理员 """
    with open(_super_user, "w+") as _su:
        sr = _su.read()
        if not sr:
            sudos = {}
        else:
            sudos = json.loads(sr)
        sudos[get_user_id(message)] = ""
        _su.write(json.dumps(sudos))
    return True

def rm_super_user(message) -> bool:
    """ 删除超级管理员 """
    rsu = open(_super_user, "r")
    sr = rsu.read()
    if not sr:
        return False
    sudos = json.loads(sr)
    try:
        sudos.pop(get_user_id(message))
    except KeyError:
        return False
    _su = open(_super_user, "w")
    _su.write(json.dumps(sudos))
    return True

def is_super_user(event) -> bool:
    """ 判断`event`所指向的用户是否为超级管理员 """
    su = False
    with open(_super_user, "r") as _su:
        sr = _su.read()
        if not sr:
            sudos = {}
        else:
            sudos = json.loads(sr)
    for sudo in sudos.keys():
        if get_user_id(event) == sudo:
            su = True
            break
    return su

def botoff(event):
    """ 机器人在`event`所指向的群聊中开启指令限制 """
    status = load_status_settings()
    status[get_group_id(event)] = False
    change_status(status)
    f = open(_dicer_girl_status, "w")
    json.dump(status, f)

def boton(event):
    """ 机器人在`event`所指向的群聊中开启完全功能 """
    status = load_status_settings()
    status[get_group_id(event)] = True
    change_status(status)
    f = open(_dicer_girl_status, "w")
    json.dump(status, f)

def get_status(event):
    """ 判断机器人在`event`所指向的群聊中是否处于完全功能状态 """
    status = load_status_settings()
    if not get_group_id(event) in status.keys():
        status[get_group_id(event)] = True
        f = open(_dicer_girl_status, "w")
        json.dump(status, f)
        return True

    return status[get_group_id(event)]

def load_status():
    """ 导入目前所存储的机器人在各群聊中状态 """
    change_status(json.load(open(_dicer_girl_status, "r")))

def rolekp(event):
    ...

def roleob(event):
    ...

async def get_latest_version(package_name):
    """ 获取当前 Pypi 上`dicergirl`的最新版本号 """
    async with httpx.AsyncClient() as client:
        url = f"https://pypi.org/pypi/{package_name}/json"
        response = await client.get(url)

        if response.status_code == 404:
            return "0.0.0.0"

        package_info = response.json()
        return package_info["info"]["version"]

async def _run_shell_command(command):
    """ 异步执行 shell 指令的原始方法 """
    process = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await process.communicate()

    return {
        "stdout": stdout.decode().strip(),
        "stderr": stderr.decode().strip(),
        "returncode": process.returncode
    }

def run_shell_command(command):
    """ 异步执行 shell 指令 """
    return asyncio.run(_run_shell_command(command))


from nonebot.adapters.onebot.v12 import GroupMessageEvent
try:
    from .docimasy import expr
    from ..utils.utils import get_group_id
    from .dicer import Dicer
    from ..utils.plugins import modes
except ImportError:
    from dicergirl.utils.docimasy import expr
    from dicergirl.utils.utils import get_group_id
    from dicergirl.utils.dicer import Dicer
    from dicergirl.utils.plugins import modes

def __set_plus_format(args: list):
    """ `.set 技能 +x`语法解析 """
    while True:
        try:
            index = args.index("+")
        except:
            break
        args[index] = args[index] + args[index+1]
        args.pop(index+1)

    while True:
        try:
            index = args.index("-")
        except:
            break
        args[index] = args[index] + args[index+1]
        args.pop(index+1)

    return args

def __set_default(args: list, event, cards=None, module=None, attrs_dict=None, cha=None, qid=None) -> bool:
    """ 技能或属性设置 """
    for attr, alias in attrs_dict.items():
        if args[0] in alias:
            if attr in ["名字", "性别"]:
                if attr == "性别" and not args[1] in ["男", "女"]:
                    return f"[Oracle] 欧若可拒绝将{module.__cname__}性别将设置为 {args[1]}, 这是对物种的侮辱."
                cha.__dict__[alias[0]] = args[1]
            else:
                try:
                    if not args[1].startswith(("-", "+")):
                        cha.__dict__[alias[0]] = int(args[1])
                    elif args[1].startswith("+"):
                        cha.__dict__[alias[0]] += int(args[1][1:])
                    elif args[1].startswith("-"):
                        cha.__dict__[alias[0]] -= int(args[1][1:])
                except ValueError:
                    return "基础数据 %s 要求正整数数据, 但你传入了 %s." % (args[0], args[1])
            cards.update(event, cha.__dict__, qid=qid)
            return "[Oracle] 设置%s %s 为: %s" % (module.__cname__, attr, cha.__dict__[alias[0]])

def __set_skill(args, event, reply: list, cards=None, cha=None, module=None, qid=None):
    """ 设置技能 """
    try:
        if not args[1].startswith(("-", "+")):
            cha.skills[args[0]] = int(args[1])
        elif args[1].startswith("+"):
            cha.skills[args[0]] += int(args[1][1:])
        elif args[1].startswith("-"):
            cha.skills[args[0]] -= int(args[1][1:])
        cards.update(event, cha.__dict__, qid=qid)
        reply.append("设置%s %s 技能为: %s." % (module.__cname__, args[0], cha.skills[args[0]]))
    except ValueError:
        reply.append("技能 %s 要求正整数数据, 但你传入了 %s." % (args[0], args[1]))
    finally:
        return reply

def set_handler(event: GroupMessageEvent, args, at, mode=None):
    """ 兼容所有模式的`.set`指令后端方法 """
    module = modes[mode]
    cards = module.__cards__
    cache_cards = module.__cache__
    charactor = module.__charactor__
    attrs_dict = module.__baseattrs__
    args = __set_plus_format(args)

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    if not args:
        if cache_cards.get(event, qid=qid):
            card_data = cache_cards.get(event, qid=qid)
            cards.update(event, inv_dict=card_data, qid=qid)
            inv = charactor().load(card_data)
            return "[Oracle] 成功从缓存保存人物卡属性: \n" + inv.output()
        else:
            return f"[Oracle] 未找到缓存数据, 请先使用无参数的`.{module.__name__}`指令进行车卡生成角色卡."
    else:
        if cards.get(event, qid=qid):
            card_data = cards.get(event, qid=qid)
            inv = charactor().load(card_data)
        else:
            return f"[Oracle] 未找到缓存数据, 请先使用无参数的`.{module.__name__}`指令进行车卡生成角色卡."

        if len(args) % 2 != 0:
            return "[Oracle] 参数错误, 这是由于传输的数据数量错误, 我只接受为偶数的参数数量.\n此外, 这看起来不像是来源于我的错误."

        elif len(args) == 2:
            sd = __set_default(args, event, cards=cards, module=module, attrs_dict=attrs_dict, cha=inv, qid=qid)
            if sd:
                return sd
            
            return __set_skill(args, event, [], cards=cards, cha=inv, module=module, qid=qid)[0]
        elif len(args) > 2:
            reply = []
            li = []
            sub_li = []
            for arg in args:
                index = args.index(arg)
                if index % 2 == 0:
                    sub_li.append(arg)
                elif index % 2 == 1:
                    sub_li.append(arg)
                    li.append(sub_li)
                    sub_li = []
                else:
                    return "[Oracle] 参数错误, 这是由于传输的数据数量错误, 我只接受为偶数的参数数量.\n此外, 这看起来不像是来源于我的错误."

            for sub_li in li:
                sd = __set_default(sub_li, event, cards=cards, module=module, attrs_dict=attrs_dict, cha=inv, qid=qid)
                if sd:
                    continue

                reply = __set_skill(sub_li, event, reply, cards=cards, cha=inv, module=module, qid=qid)

            rep = "[Oracle]\n"
            for r in reply:
                rep += r + "\n"
            return rep.rstrip("\n")
        else:
            return "[Oracle] 参数错误, 可能是由于传输的数据数量错误.\n此外, 这看起来不像是来源于我的错误."

def show_handler(message, args, at, mode=None):
    """ 兼容所有模式的`.show`指令后端方法 """
    module = modes[mode]
    cards = module.__cards__
    cache_cards = module.__cache__
    charactor = module.__charactor__

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    r = []
    if not args:
        if cards.get(message, qid=qid):
            card_data = cards.get(message, qid=qid)
            inv = charactor().load(card_data)
            data = "[Oracle] 使用中人物卡: \n" 
            data += inv.output()
            r.append(data)
        if cache_cards.get(message, qid=qid):
            card_data = cache_cards.get(message, qid=qid)
            inv = charactor().load(card_data)
            r.append("[Oracle] 已暂存人物卡: \n" + inv.output())
    elif args[0] in ["detail", "de", "details"]:
        if cards.get(message, qid=qid):
            card_data = cards.get(message, qid=qid)
            inv = charactor().load(card_data)
            r.append(inv.skills_output())
    elif args[0] == "all":
        cd = cards.data[get_group_id(message)]
        for data in cd:
            inv = charactor().load(cd[data])
            d = inv.output() + "\n"
            d += inv.skills_output()
            r.append(d)
    else:
        if cards.get(message, qid=qid):
            card_data = cards.get(message, qid=qid)
            cha = charactor().load(card_data)
            try:
                r.append(getattr(cha, "out_"+args[0])())
            except:
                r.append("[Oracle] 查询时出现异常, 可能你想要查询的内容不存在?")

    if not r:
        r.append("[Oracle] 未查询到保存或暂存信息.")

    return r

def del_handler(message, args, at, mode=None):
    """ 兼容所有模式的`.del`指令后端方法 """
    module = modes[mode]
    cache_cards = module.__cache__
    cards = module.__cards__

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    r = []
    for arg in args:
        if not arg:
            pass
        elif arg == "cache":
            if cache_cards.get(message, qid=qid):
                if cache_cards.delete(message, save=False):
                    r.append("[Oracle] 已清空暂存人物卡数据.")
                else:
                    r.append("[Oracle] 错误: 未知错误.")
            else:
                r.append("[Oracle] 暂无缓存人物卡数据.")
        elif arg == "card":
            if cards.get(message):
                if cards.delete(message):
                    r.append("[Oracle] 已删除使用中的人物卡！")
                else:
                    r.append("[Oracle] 错误: 未知错误.")
            else:
                r.append("[Oracle] 暂无使用中的人物卡.")
        else:
            if cards.delete_skill(message, arg):
                r.append(f"已删除技能 {arg}.")

    if not r:
        r.append("[Oracle] 使用`.help del`获得指令使用帮助.")

    return r

def roll(args: str, name: str=None) -> str:
    """ 标准掷骰指令后端方法 """
    time = 1
    if "#" in args:
        args = args.split("#")

        try:
            time = int(args[0].strip())
        except ValueError:
            return "[Oracle] 参数错误, `#`提示符前应当跟随整型数."

        if len(args) == 1:
            args = "1d100"
        else:
            args = args[1]
    else:
        args = args.strip()

    try:
        d = Dicer(args)
        r = expr(d, None, name=name)

        for _ in range(time-1):
            r += expr(d, None, name=name)

        return r.detail
    except ValueError:
        return "[Oracle] 出现错误, 请检查你的掷骰表达式.\n使用`.help r`获得掷骰指令使用帮助."

def shoot():
    dice = Dicer("1d20").roll()
    result = dice.outcome

    if result < 4:
        rstr = "右腿"
    elif result < 7:
        rstr = "左腿"
    elif result < 11:
        rstr = "腹部"
    elif result < 16:
        rstr = "胸部"
    elif result < 18:
        rstr = "右臂"
    elif result < 20:
        rstr = "左臂"
    elif result < 21:
        rstr = "头部"

    return f"[Oracle] 进行射击检定:\n{dice.description()}\n命中了 {rstr}."


try:
    from ..utils.dicer import Dicer
    from .attributes import scp_attrs_dict, ability_data, knowledge_data, skills_data
except ImportError:
    from dicergirl.utils.dicer import Dicer
    from dicergirl.scp.attributes import scp_attrs_dict, knowledge_data, skills_data, ability_data

import random

class Agent(object):
    """ SCP 特工人物卡 """
    def __init__(self) -> None:
        self.name = "无名特工"
        self.age = 20
        self.sex = "女"
        self.level = 1
        self.hp_max = 0
        self.hp = 0
        self.enp = 0
        self.rep = 0
        self.ach = 0
        self.dices = {}
        self.en = {}
        self.money = 200
        self.agentclass = "E"
        self.knowledge = {knowledge: 0 for knowledge in knowledge_data.keys()}
        self.skills = {skill: 0 for skill in skills_data.keys()}
        self.ability = {ability: 0 for ability in ability_data.keys()}
        self.p = {
            "knowledge": 17,
            "skills": 14,
            "ability": 10
        }
        self.ix = 1
        self.rx = 1
        self.tools = {}

    def init(self):
        prop = {
            "str": 1,
            "hth": 1,
            "per": 1,
            "dex": 1,
            "fte": 1,
            "chr": 1,
            "int": 1,
            "wil": 1
        }
        attr = {p: 1 for p in prop}
        self.en = {p: 0 for p in prop}
        total = 20 - len(prop)
        for _ in range(total):
            name = random.choice(list(prop.keys()))
            prop[name] += 1

        for p in prop.keys():
            num = prop[p]
            dice = Dicer(f"{num}d8", explode=True)
            dice.roll()
            prop[p] = dice.outcome
            attr[p] = dice.dices

        self.dices = attr
        self.reset()

    def reset(self):
        self.reset_hp()
        self.reset_enp()
        self.reset_p()
        self.reset_rep()
        self.reset_x()

    def reset_x(self):
        self.ix = 1
        self.rx = 1
        for d in self.dices["str"]:
            if d == "D10":
                self.ix += 1
            elif d == "D12":
                self.ix += 1

        for d in self.dices["per"]:
            if d == "D10":
                self.rx += 1
            elif d == "D12":
                self.rx += 1

    def reset_card(self):
        self.reset()
        self.money = 200
        self.agentclass = "E"
        self.knowledge = {knowledge: 0 for knowledge in knowledge_data.keys()}
        self.skills = {skill: 0 for skill in skills_data.keys()}
        self.ability = {ability: 0 for ability in ability_data.keys()}
        self.p = {
            "knowledge": 17,
            "skills": 14,
            "ability": 10
        }
        self.tools = {}
        self.ach = 0
        self.level = 1

    def reset_hp(self):
        base = 10
        for d in self.dices["hth"]:
            if d == "D8":
                base += 3
            elif d == "D10":
                base += 6
            elif d == "D12":
                base += 16
        self.hp_max = base
        self.hp = base
    
    def reset_enp(self):
        base = 1
        for d in self.dices["wil"]:
            if d == "D10":
                base += 1
        self.enp = base
    
    def reset_p(self):
        self.p = {
            "knowledge": 17,
            "skills": 14,
            "ability": 10
        }
        self.p["knowledge"] += (self.__count("D10", self.dices["int"]) + self.__count("D10", self.dices["per"])) * 5
        self.p["skills"] += (self.__count("D10", self.dices["str"]) + self.__count("D10", self.dices["dex"])) * 5
        self.p["ability"] += (self.__count("D10", self.dices["chr"]) + self.__count("D10", self.dices["wil"])) * 5
    
    def reset_rep(self):
        self.rep = 0
        for i in self.dices["fte"]:
            if i == "D8":
                self.rep += 1
            elif i == "D10":
                self.rep += 3

    def __count(self, string, list):
        count = 0
        for item in list:
            if string == item:
                count += 1
        return count

    def age_check(self, age=20):
        if self.age != 20:
            return
        if age < 15:
            return "[scpdicer] 年龄过小, 无法担任基金会特工."
        elif age >= 90:
            return "该特工已经被清理, 他(或者她)年龄过大, 显然是一个需要被清理的异常."
        self.age = age

    def __repr__(self) -> str:
        data = "姓名: %s\n" % self.name
        data += "性别: %s\n年龄: %d\n" % (self.sex, self.age)
        data += "强度: %s\n命运: %s\n" % (self.__dices_format("str"), self.__dices_format("fte"))
        data += "感知: %s\n魅力: %s\n" % (self.__dices_format("per"), self.__dices_format("chr")) 
        data += "灵巧: %s\n情报: %s\n" % (self.__dices_format("dex"), self.__dices_format("int"))
        data += "健康: %s\n意志: %s\n" % (self.__dices_format("hth"), self.__dices_format("wil"))
        data += "熟练值:\n"
        data += "  知识: %s " % self.p["knowledge"]
        data += "技能: %s " % self.p["skills"]
        data += "能力: %s\n" % self.p["ability"]
        data += "声望: %s\n" % (self.rep)
        data += "激励点: %d\n" % (self.enp)
        data += "生命: %d/%d" % (self.hp, self.hp_max)
        return data
    
    def __dices_format(self, prop):
        d8 = 0
        d10 = 0
        d12 = 0
        d20 = 0

        for dice in self.dices[prop]:
            if dice == "D8":
                d8 += 1
            elif dice == "D10":
                d10 += 1
            elif dice == "D12":
                d12 += 1
            elif dice == "D20":
                d20 += 1

        d8 = f"{d8}D8"
        d10 = f"+{d10}D10" if d10 != 0 else ""
        d12 = f"+{d12}D12" if d12 != 0 else ""
        d20 = f"+{d20}D20" if d20 != 0 else ""
        return d8 + d10 + d12 + d20

    def skills_output(self) -> str:
        if not self.skills and not self.knowledge and not self.ability:
            return "%s 当前无任何技能数据。" % self.name

        r = self.__skill_output_format("知识", self.knowledge.items()) + "\n"
        r += "\n" + self.__skill_output_format("技巧", self.skills.items()) + "\n"
        r += "\n" + self.__skill_output_format("能力", self.ability.items())

        return r

    def __skill_output_format(self, name, items):
        r = f"{self.name} {name}"
        count = 0
        for k, v in items:
            if count % 2 == 0:
                line = "\n"
                tab = " "
            else:
                line = ""
                tab = ""

            r += f"{line}{k}: {v}{tab}"
            count += 1
        return r

    def output(self) -> str:
        return self.__repr__()
    
    def all_not_base(self):
        nbs = {k: "skills" for k, v in self.skills.items()}
        nbk = {k: "knowledge" for k, v in self.knowledge.items()}
        nba = {k: "ability" for k, v in self.ability.items()}
        anb = {}
        anb.update(nbs)
        anb.update(nbk)
        anb.update(nba)
        return anb

    def out_skills(self):
        return self.skills_output()

    def out_x(self):
        return f"角色近战系数: {self.ix}\n角色远程系数: {self.rx}"

    def out_knowledge(self):
        return self.__skill_output_format("知识", self.knowledge.items()).strip("\n") if self.knowledge else "%s 当前无任何知识数据." % self.name

    def out_skills(self):
        return self.__skill_output_format("技能", self.skills.items()).strip("\n") if self.skills else "%s 当前无任何能力数据." % self.name

    def out_ability(self):
        return self.__skill_output_format("能力", self.ability.items()).strip("\n") if self.ability else "%s 当前无任何能力数据." % self.name

    def out_hp(self):
        return "当前生命值: %d/%d" % (self.hp, self.hp_max)

    def out_enp(self):
        return f"当前剩余激励点: {self.enp}"

    def out_rep(self):
        return f"当前声望: {self.rep}"

    def out_p(self):
        out =  "当前熟练值:\n"
        out += "  知识: %d" % self.p["knowledge"]
        out += "  技能: %d" % self.p["skills"]
        out += "  能力: %d" % self.p["ability"]
        return out

    def out_exp(self):
        return self.out_p()

    def out_money(self):
        return f"当前余额: {self.money}"

    def out_class(self):
        return f"当前特工类别: {self.agentclass}"

    def out_level(self):
        return f"当前特工权限: {self.level}"

    def out_en(self):
        out = "当前激励状态:"
        for en in self.en.keys():
            ability = None
            for attr in scp_attrs_dict.keys():
                if scp_attrs_dict[attr][0] == en:
                    ability = attr

            if not ability:
                return f"[Orcale] 错误: 已经被激励的技能 {en} 不存在!"

            out += f"\n  {ability}: {self.en[en]} 点激励."

        return out

    def load(self, data: dict):
        self.__dict__.update(data)
        return self